@@ -112,19 +112,19 @@ struct nvec_subdev {
 };
 
 /**
- * struct nvec_platform_data - Platform data for NVIDIA Embedded Contoller
- * @gpio: GPIO number of the EC request pin
- * @adapter: number of the slave controller
- * @custom_drivers: board specific drivers like special events or device
+ * struct nvec_platform_data - platform data for a tegra slave controller
+ * @i2c_addr: number of i2c slave adapter the ec is connected to
+ * @gpio: gpio number for the ec request line
+ * @nvec_devices: board specific drivers like special events or device
  *                  initializations
- * @nr_custom_devs: number of entries in the custom_devices array
+ * @nr_nvec_devs: number of entries in the custom_devices array
  *
  * Platform data, to be used in board definitions. For an example, take a
  * look at the paz00 board in arch/arm/mach-tegra/board-paz00.c
  */
 struct nvec_platform_data {
+	int i2c_addr;
 	int gpio;
-	int adapter;
 	struct mfd_cell *nvec_devices;
 	int nr_nvec_devs;
 	bool has_poweroff;
@@ -135,9 +135,9 @@ struct nvec_platform_data {
  * @dev: The device
  * @gpio: GPIO number of the EC request pin
  * @irq: The IRQ of the I2C device
- * @custom_devices: board specific drivers like special events or device
+ * @nvec_devices: board specific drivers like special events or device
  *                  initializations
- * @nr_custom_devs: number of entries in the custom_devices array
+ * @nr_nvec_devs: number of entries in the custom_devices array
  * @i2c_addr: The address of the I2C slave
  * @base: The base of the memory mapped region of the I2C device
  * @clk: The clock of the I2C device
@@ -149,8 +149,8 @@ struct nvec_platform_data {
  * @rx_work: A work structure for the RX worker nvec_dispatch()
  * @tx_work: A work structure for the TX worker nvec_request_master()
  * @wq: The work queue in which @rx_work and @tx_work are executed
- * @msg_pool: A pool of messages for allocation
  * @rx: The message currently being retrieved or %NULL
+ * @msg_pool: A pool of messages for allocation
  * @tx: The message currently being transferred
  * @tx_scratch: Used for building pseudo messages
  * @ec_transfer: A completion that will be completed once a message has been
@@ -171,8 +171,7 @@ struct nvec_chip {
 	int nr_nvec_devs;
 	int i2c_addr;
 	void __iomem *base;
-	struct clk *clk;
-
+	struct clk *i2c_clk;
 	struct atomic_notifier_head notifier_list;
 	struct list_head rx_data, tx_data;
 	struct notifier_block nvec_status_notifier;
